@@ -9,6 +9,9 @@ from einops import rearrange
 def exists(val):
     return val is not None
 
+def default(val, d):
+    return val if exists(val) else d
+
 # classes
 
 class Residual(nn.Module):
@@ -30,12 +33,17 @@ class PreNorm(nn.Module):
 
 # attention
 
+class GEGLU(nn.Module):
+    def forward(self, x):
+        x, gates = x.chunk(2, dim = -1)
+        return x * F.gelu(gates)
+
 class FeedForward(nn.Module):
     def __init__(self, dim, mult = 4):
         super().__init__()
         self.net = nn.Sequential(
-            nn.Linear(dim, dim * mult),
-            nn.GELU(),
+            nn.Linear(dim, dim * mult * 2),
+            GEGLU(),
             nn.Linear(dim * mult, dim)
         )
 
@@ -69,6 +77,29 @@ class Attention(nn.Module):
         out = rearrange(out, 'b h n d -> b n (h d)', h = h)
         return self.to_out(out)
 
+# mlp
+
+class MLP(nn.Module):
+    def __init__(self, dims, act = None):
+        super().__init__()
+        dims_pairs = list(zip(dims[:-1], dims[1:]))
+        layers = []
+        for ind, (dim_in, dim_out) in enumerate(dims_pairs):
+            is_last = ind >= (len(dims) - 1)
+            linear = nn.Linear(dim_in, dim_out)
+            layers.append(linear)
+
+            if is_last:
+                continue
+
+            act = default(act, nn.ReLU())
+            layers.append(act)
+
+        self.mlp = nn.Sequential(*layers)
+
+    def forward(self, x):
+        return self.mlp(x)
+
 # main class
 
 class TabTransformer(nn.Module):
@@ -83,28 +114,39 @@ class TabTransformer(nn.Module):
         dim_head = 16,
         dim_out = 1,
         mlp_hidden_mults = (4, 2),
-        continuous_mean_var = None
+        mlp_act = None,
+        num_special_tokens = 2,
+        continuous_mean_std = None
     ):
         super().__init__()
         assert all(map(lambda n: n > 0, categories)), 'number of each category must be positive'
-        assert len(mlp_hidden_mults) == 2, 'final mlp fixed at 2 layers for now'
 
         # categories related calculations
 
         self.num_categories = len(categories)
         self.num_unique_categories = sum(categories)
 
+        # create category embeddings table
+
+        self.num_special_tokens = num_special_tokens
+        total_tokens = self.num_unique_categories + num_special_tokens
+
+        self.categorical_embeds = nn.Embedding(total_tokens, dim)
+
         # for automatically offsetting unique category ids to the correct position in the categories embedding table
 
-        categories_offset = F.pad(torch.tensor(list(categories)), (1, 0), value = 0).cumsum(dim = -1)[:-1]
+        categories_offset = F.pad(torch.tensor(list(categories)), (1, 0), value = num_special_tokens)
+        categories_offset = categories_offset.cumsum(dim = -1)[:-1]
         self.register_buffer('categories_offset', categories_offset)
 
-        self.categorical_embeds = nn.Embedding(self.num_unique_categories, dim)
-
         # continuous
 
-        assert continuous_mean_var.shape == (num_continuous, 2), f'continuous_mean_var must have a shape of ({num_continuous}, 2) where the last dimension contains the mean and variance respectively'
-        self.register_buffer('continuous_mean_var', continuous_mean_var)
+        if exists(continuous_mean_std):
+            assert continuous_mean_std.shape == (num_continuous, 2), f'continuous_mean_std must have a shape of ({num_continuous}, 2) where the last dimension contains the mean and variance respectively'
+            self.register_buffer('continuous_mean_std', continuous_mean_std)
+
+        self.norm = nn.LayerNorm(num_continuous)
+        self.num_continuous = num_continuous
 
         # attention layers
 
@@ -116,21 +158,15 @@ class TabTransformer(nn.Module):
                 Residual(PreNorm(dim, FeedForward(dim))),
             ]))
 
-        self.norm = nn.LayerNorm(num_continuous)
-
         # mlp to logits
 
         input_size = (dim * self.num_categories) + num_continuous
         l = input_size // 8
-        mult1, mult2 = mlp_hidden_mults
 
-        self.mlp = nn.Sequential(
-            nn.Linear(input_size, l * mult1),
-            nn.ReLU(),
-            nn.Linear(l * mult1, l * mult2),
-            nn.ReLU(),
-            nn.Linear(l * mult2, dim_out)
-        )
+        hidden_dimensions = list(map(lambda t: l * t, mlp_hidden_mults))
+        all_dimensions = [input_size, *hidden_dimensions, dim_out]
+
+        self.mlp = MLP(all_dimensions, act = mlp_act)
 
     def forward(self, x_categ, x_cont):
         assert x_categ.shape[-1] == self.num_categories, f'you must pass in {self.num_categories} values for your categories input'
@@ -144,9 +180,11 @@ class TabTransformer(nn.Module):
 
         flat_categ = x.flatten(1)
 
-        if exists(self.continuous_mean_var):
-            mean, var = self.continuous_mean_var.unbind(dim = -1)
-            x_cont = (x_cont - mean) / var
+        assert x_cont.shape[1] == self.num_continuous, f'you must pass in {self.num_continuous} values for your continuous input'
+
+        if exists(self.continuous_mean_std):
+            mean, std = self.continuous_mean_std.unbind(dim = -1)
+            x_cont = (x_cont - mean) / std
 
         normed_cont = self.norm(x_cont)
 
